#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from eth_typing import ChecksumAddress
from web3.contract.contract import ContractFunction

from skale.contracts.base_contract import transaction_method
from skale.contracts.ima.schain.base_token_manager import BaseTokenManager
from skale.types.schain import SchainName
from skale.utils.helper import schain_name_to_hash


class TokenManagerERC721WithMetadata(BaseTokenManager):
    @transaction_method
    def add_erc721(self, schain_name: SchainName, token_mn: int, token_sc: int) -> ContractFunction:
        return self.contract.functions.addERC721TokenByOwner(schain_name, token_mn, token_sc)

    @transaction_method
    def transfer_to_schain_erc721(
        self, schain_name: SchainName, address: ChecksumAddress, token_id: int
    ) -> ContractFunction:
        """address - token address on origin chain"""
        return self.contract.functions.transferToSchainERC721(schain_name, address, token_id)

    @transaction_method
    def exit_to_main_erc721(self, address: ChecksumAddress, token_id: int) -> ContractFunction:
        return self.contract.functions.exitToMainERC721(address, token_id)

    def get_clones_erc721(self, schain_name: SchainName, address: ChecksumAddress) -> int:
        return self.contract.functions.clonesErc721(
            schain_name_to_hash(schain_name), address
        ).call()



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from eth_typing import ChecksumAddress
from web3.contract.contract import ContractFunction

from skale.contracts.base_contract import transaction_method
from skale.contracts.skale_contract import SkaleContract


class Context(SkaleContract):
    """Context contract"""

    def get_schain_name(self) -> str:
        return self.contract.functions.getSchainName().call()

    def get_schain_owner_address(self) -> bytes:
        return self.contract.functions.getSchainOwnerAddress().call()

    @transaction_method
    def set_schain_owner_address(self, newOwner: ChecksumAddress) -> ContractFunction:
        return self.contract.functions.setSchainOwnerAddress(newOwner)

    def get_version(self) -> str:
        return self.contract.functions.version().call()



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from eth_typing import ChecksumAddress
from web3.contract.contract import ContractFunction

from skale.contracts.base_contract import transaction_method
from skale.contracts.skale_contract import SkaleContract
from skale.types.schain import SchainName


class MessageProxyForMainnet(SkaleContract):
    @transaction_method
    def register_extra_contract(
        self, schain_name: SchainName, contract_address: ChecksumAddress
    ) -> ContractFunction:
        return self.contract.functions.registerExtraContract(schain_name, contract_address)

    @transaction_method
    def remove_extra_contract(
        self, schain_name: SchainName, contract_address: ChecksumAddress
    ) -> ContractFunction:
        return self.contract.functions.removeExtraContract(schain_name, contract_address)

    @transaction_method
    def add_reimbursed_contract(
        self, schain_name: SchainName, contract_address: ChecksumAddress
    ) -> ContractFunction:
        return self.contract.functions.addReimbursedContract(schain_name, contract_address)

    @transaction_method
    def remove_reimbursed_contracts(
        self, schain_name: SchainName, contract_address: ChecksumAddress
    ) -> ContractFunction:
        return self.contract.functions.removeReimbursedContract(schain_name, contract_address)

    @transaction_method
    def pause(self, schain_name: SchainName) -> ContractFunction:
        return self.contract.functions.pause(schain_name)

    @transaction_method
    def resume(self, schain_name: SchainName) -> ContractFunction:
        return self.contract.functions.resume(schain_name)

    def is_connected_chain(self, schain_name: SchainName) -> bool:
        return self.contract.functions.isConnectedChain(schain_name).call()

    def is_reimbursed_contract(self, schain_name: SchainName) -> bool:
        return self.contract.functions.isReimbursedContract(schain_name).call()

    def is_paused(self, schian_name: str) -> bool:
        return self.contract.functions.isPaused(schian_name).call()

    def extra_contract_register_role(self) -> bytes:
        return self.contract.functions.EXTRA_CONTRACT_REGISTRAR_ROLE().call()

    def pausable_role(self) -> bytes:
        return self.contract.functions.PAUSABLE_ROLE().call()

    @transaction_method
    def grant_role(self, role: bytes, address: ChecksumAddress) -> ContractFunction:
        return self.contract.functions.grantRole(role, address)

    def has_role(self, role: bytes, address: ChecksumAddress) -> bool:
        return self.contract.functions.hasRole(role, address).call()

    def get_role_member(self, role: bytes, index: int) -> bytes:
        return self.contract.functions.getRoleMember(role, index).call()



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.


import abc
import logging

import requests
from skale_contracts import skale_contracts
from skale_contracts.project_factory import SkaleProject
from web3 import Web3

from skale.rpc_session import CountingSession, RpcHttpStats
from skale.utils.cache import RedisCacheConfig
from skale.utils.exceptions import EmptyWalletError, InvalidWalletError
from skale.utils.helper import contract_name_to_snake_case
from skale.utils.web3_utils import default_gas_price, get_endpoint, init_web3
from skale.wallets import BaseWallet

logger = logging.getLogger(__name__)


class EmptyPrivateKey(Exception):
    pass


class SkaleBase:
    __metaclass__ = abc.ABCMeta

    DEFAULT_RPC_CACHE_TTL_POLICY: dict[str, int] = {
        'eth_call': 10,
        'eth_getCode': 30,
        'eth_getStorageAt': 30,
        'eth_chainId': 600,
        'eth_getBlockByNumber': 60,
        'eth_gasPrice': 5,
        'web3_clientVersion': 600,
    }

    def __init__(
        self,
        endpoint: str | list[str],
        alias_or_address: str,
        wallet: BaseWallet | None = None,
        state_path: str | None = None,
        ts_diff: int | None = None,
        provider_timeout: int = 30,
        session: requests.Session | None = None,
        enable_stats: bool = False,
        redis_cache_url: str | None = None,
        rpc_cache_methods: list[str] | None = None,
        rpc_cache_bypass_addresses: list[str] | None = None,
        rpc_cache_ttl_policy: dict[str, int] | None = None,
    ):
        logger.info(
            'Initializing %s, endpoint: %s, alias_or_address: %s, wallet: %s',
            self.__class__.__name__,
            endpoint,
            alias_or_address,
            type(wallet).__name__,
        )
        if state_path:
            logger.warning(
                'state_path is deprecated and will be ignored. This option will be removed in v8.'
            )
        self.stats = RpcHttpStats()
        if session is None and enable_stats:
            session = CountingSession(self.stats)
        self._endpoint = get_endpoint(endpoint, ts_diff=ts_diff, provider_timeout=provider_timeout)
        self._alias_or_address = alias_or_address

        cache_config = None
        if redis_cache_url is not None:
            cache_methods = rpc_cache_methods or [
                'eth_call',
                'eth_getCode',
                'eth_getStorageAt',
                'eth_chainId',
                'eth_getBlockByNumber',
                'eth_gasPrice',
                'web3_clientVersion',
            ]
            cache_ttl_policy = rpc_cache_ttl_policy or self.DEFAULT_RPC_CACHE_TTL_POLICY
            cache_config = RedisCacheConfig(
                redis_url=redis_cache_url,
                cache_methods=frozenset(cache_methods),
                bypass_addresses=frozenset(rpc_cache_bypass_addresses or []),
                default_ttl_seconds=0,
                method_ttl_policy=cache_ttl_policy,
            )
        self.web3 = init_web3(
            self._endpoint,
            ts_diff=ts_diff,
            provider_timeout=provider_timeout,
            session=session,
            cache_config=cache_config,
        )
        self.network = skale_contracts.get_network_by_provider(self.web3.provider)
        self.project = self.network.get_project(self.project_name)
        self.instance = self.project.get_instance(alias_or_address)
        if wallet:
            self.wallet = wallet

    @property
    @abc.abstractmethod
    def project_name(self) -> SkaleProject:
        """Name of smart contracts project"""

    @property
    def gas_price(self) -> int:
        return default_gas_price(self.web3)

    @property
    def wallet(self) -> BaseWallet:
        if not self._wallet:
            raise EmptyWalletError('No wallet provided')
        return self._wallet

    @wallet.setter
    def wallet(self, wallet: BaseWallet) -> None:
        if issubclass(type(wallet), BaseWallet):
            self._wallet = wallet
        else:
            raise InvalidWalletError(
                f'Wrong wallet class: {type(wallet).__name__}. \
                Must be one of the BaseWallet subclasses'
            )

    def _generate_legacy_abi(self) -> dict:
        abi = {}
        for contract_name in self.instance.abi.keys():
            if contract_name == 'RewardWallet':  # RewardWallet requires node id to get address
                continue
            address = Web3.to_checksum_address(self.instance.get_contract_address(contract_name))
            name = contract_name_to_snake_case(contract_name)
            abi[f'{name}_abi'] = self.instance.abi[contract_name]
            abi[f'{name}_address'] = address
        return abi
